'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var deepExtend = require('deep-extend');
var jsonpath = require('jsonpath');
var diff = require('simple-diff');
var expressionExpander = require('expression-expander');
var micromatch = require('micromatch');

/**
 * Mergable File
 * @param {string} path location in the repository
 * @param {Object} options mergin options
 *
 * @property {string} path
 * @property {Object} options
 */
class File {
  static matchesFileName(name) {
    return false;
  }

  static get defaultOptions() {
    return {};
  }

  constructor(path, options = {}) {
    Object.defineProperties(this, {
      path: {
        value: path
      },
      options: {
        value: Object.assign({}, this.defaultOptions, options)
      }
    });
  }

  get defaultOptions() {
    return this.constructor.defaultOptions;
  }

  get needsTemplate() {
    return true;
  }

  get needsOriginal() {
    return false;
  }

  optionalDevModules(modules) {
    return new Set();
  }

  usedDevModules(content) {
    return new Set();
  }

  /**
   * Deliver some key properties
   * @param {Branch} branch
   * @return {Object}
   */
  async properties(branch) {
    return {};
  }

  async targetContent(context, options) {
    return (await context.targetBranch.content(this.path, options)).content;
  }

  async content(context) {
    return (await Promise.all([
      context.targetBranch.content(this.path, {
        ignoreMissing: !this.needsOriginal
      }),
      context.templateBranch.content(this.path, {
        ignoreMissing: !this.needsTemplate
      })
    ])).map(c => c.content);
  }

  async mergeContent(context, original, template) {
    return {
      changed: false,
      content: original
    };
  }

  /**
   * @param {PreparedContect} context
   * @return {Object} merged content
   */
  async merge(context) {
    try {
      const [original, template] = await this.content(context);
      const result = await this.mergeContent(context, original, template);
      if (result === undefined) {
        return {
          path: this.path,
          changed: false
        };
      }
      result.path = this.path;

      context.properties.path = this.path;
      result.messages = context.expand(result.messages);

      return result;
    } catch (err) {
      context.error(
        `${
          context.targetBranch
            ? context.targetBranch.fullCondensedName
            : 'unknown'
        },${this.path}: ${err}`
      );
      return {
        path: this.path,
        changed: false
      };
    }
  }
}

/**
 * find merger options in the template section of a package.json
 * @param {Object} json
 * @param {string} name
 * @return {Object}
 */
function templateOptions(json, name) {
  if (json.template !== undefined && json.template.files !== undefined) {
    const m = json.template.files.find(f => f.merger === name);
    if (m !== undefined && m.options !== undefined) {
      return m.options;
    }
  }
  return {};
}

/**
 * @param {string} a
 * @param {string} b
 */
function diffVersion(a, b) {
  const aa = String(a)
    .split(/\./)
    .map(x => parseInt(x, 10));
  const bb = String(b)
    .split(/\./)
    .map(x => parseInt(x, 10));

  for (const i in aa) {
    if (i >= bb.length) {
      break;
    }

    if (aa[i] < bb[i]) {
      return -1;
    }
    if (aa[i] > bb[i]) {
      return 1;
    }
  }

  return 0;
}

const yaml = require('js-yaml');

function difference(a, b) {
  return new Set([...a].filter(x => !b.has(x)));
}

class Travis extends File {
  static matchesFileName(name) {
    return name === '.travis.yml';
  }

  async mergeContent(context, original, template) {
    const ymlOptions = { schema: yaml.FAILSAFE_SCHEMA };

    const yml = yaml.safeLoad(original, ymlOptions) || {};
    const tyml = yaml.safeLoad(context.expand(template), ymlOptions);
    const before_script = yml.before_script;
    const email = yml.notifications ? yml.notifications.email : undefined;
    const messages = [];

    const oldVersions = new Set(
      yml.node_js ? [...yml.node_js.map(s => String(s))] : []
    );
    const templateVersions = new Set(
      tyml.node_js ? [...tyml.node_js.map(s => String(s))] : []
    );
    const versions = new Set([...oldVersions, ...templateVersions]);
    const newVersions = new Set(versions);

    versions.forEach(v => {
      if (v.startsWith('-')) {
        const d = v.replace(/^\-\s*/, '');

        versions.forEach(v => {
          const x = v.replace(/^\-\s*/, '');
          //console.log(`${d}<>${v} => ${diffVersion(d, x) === 0 || x != v}`);
          if (diffVersion(d, x) === 0 || x != v) {
            if (templateVersions.has(x)) {
              return;
            }

            newVersions.delete(x);
            newVersions.delete(v);
          }
        });
      }
    });

    const r = difference(oldVersions, newVersions);
    if (r.size > 0) {
      messages.push(
        `chore(travis): remove node versions ${Array.from(new Set(r)).sort()}`
      );
    }

    const a = difference(newVersions, oldVersions);
    if (a.size > 0) {
      messages.push(
        `chore(travis): add node versions ${Array.from(new Set(a)).sort()}`
      );
    }

    deepExtend(yml, tyml);

    if (newVersions.size > 0) {
      yml.node_js = Array.from(new Set(newVersions))
        .sort()
        .map(s => (String(parseFloat(s)) == s ? parseFloat(s) : s));
    }

    if (email !== undefined) {
      yml.notifications.email = email;
    }

    if (before_script !== undefined) {
      before_script.forEach(s => {
        if (!yml.before_script.find(e => e === s)) {
          yml.before_script.push(s);
        }
      });
    }

    if (yml.before_script) {
      yml.before_script = yml.before_script.filter(
        s =>
          !tyml.before_script.find(e => e === `-${s}`) && s.indexOf('-') !== 0
      );
    }

    const content = yaml.safeDump(yml, {
      lineWidth: 128
    });

    if (messages.length === 0) {
      messages.push(`chore(travis): merge from template ${this.path}`);
    }

    return {
      content,
      changed: content !== original,
      messages
    };
  }
}

/**
 * injects badges into REAMDE.md
 */
class Readme extends File {
  static matchesFileName(name) {
    return name.match(/README\./);
  }

  static get defaultOptions() {
    return {
      badges: []
    };
  }

  async mergeContent(context, original, template) {
    const [pkg, pkgTemplate] = await context.files
      .get('package.json')
      .content(context);

    const p = pkg.length === 0 ? {} : JSON.parse(pkg);
    const pTemplate = JSON.parse(pkgTemplate);

    const badges = this.options.badges
      .map(b => {
        const m = templateOptions(p, 'Readme');

        // TODO do not alter global properties use private layer here
        if (m.badges !== undefined) {
          Object.assign(context.properties, m.badges[b.name]);
        }

        const r = context.expand(`[![${b.name}](${b.icon})](${b.url})`);

        if (r.match(/\{\{/)) {
          return '';
        }
        return r;
      })
      .filter(b => b.length > 0);

    let body = original.split(/\n/);

    if (body.length === 0) {
      body = context.expand(template).split(/\n/);
    } else {
      body = body.filter(l => !l.match(/^\[\!\[.*\)$/));
    }

    const content = [...badges, ...body].join('\n');
    return {
      content,
      changed: content !== original,
      messages: ['docs(README): update from template']
    };
  }
}

function moduleNames(object) {
  if (object === undefined) return new Set();

  const modules = new Set();

  Object.keys(object).forEach(k => {
    const v = object[k];
    if (typeof v === 'string') {
      modules.add(v);
    } else if (Array.isArray(v)) {
      v.forEach(e => {
        if (typeof e === 'string') {
          modules.add(e);
        }
      });
    }
  });

  return modules;
}

/**
 * order in which json keys are written
 */
const sortedKeys = [
  'name',
  'version',
  'private',
  'publishConfig',
  'main',
  'browser',
  'module',
  'description',
  'keywords',
  'author',
  'maintainers',
  'contributors',
  'license',
  'bin',
  'scripts',
  'dependencies',
  'devDependencies',
  'peerDependencies',
  'optionalDependencies',
  'bundledDependencies',
  'engines',
  'os',
  'arch',
  'repository',
  'directories',
  'files',
  'man',
  'bugs',
  'homepage',
  'release',
  'ava',
  'nyc',
  'xo',
  'template'
];

/**
 * Merger for package.json
 */
class Package extends File {
  static decodeScripts(scripts) {
    if (scripts === undefined) {
      return undefined;
    }

    const decoded = {};

    Object.keys(scripts).forEach(key => {
      const script = scripts[key];

      //console.log(`${key} ${JSON.stringify(script)}`);
      if (script.match(/&&/)) {
        decoded[key] = { op: '&&', args: script.split(/\s*&&\s*/) };
      } else {
        decoded[key] = { value: script };
      }
    });

    return decoded;
  }

  static mergeScripts(source, dest) {
    if (source === undefined) {
      if (dest === undefined) {
        return undefined;
      }
      source = {};
    } else if (dest === undefined) {
      dest = {};
    }

    function mergeOP(a, b) {
      const args = x => {
        if (x === undefined) return [];
        return x.args === undefined ? [x.value] : x.args;
      };

      const t = args(a).concat(args(b));

      return {
        op: '&&',
        args: t.filter((item, pos) => t.indexOf(item) == pos)
      };
    }

    Object.keys(source).forEach(key => {
      let d = dest[key];

      const s = source[key];
      switch (s.op) {
        case '&&':
          d = mergeOP(s, d);
          break;

        default:
          if (d === undefined) {
            d = { value: s.value };
          } else {
            if (d.op === '&&') {
              d = mergeOP(s, d);
            } else {
              d.value = s.value;
            }
          }
      }

      dest[key] = d;
    });

    return dest;
  }

  static encodeScripts(encoded) {
    if (encoded === undefined) {
      return undefined;
    }

    const scripts = {};

    Object.keys(encoded).forEach(key => {
      const e = encoded[key];
      switch (e.op) {
        case '&&':
          scripts[key] = e.args.join(' && ');
          break;

        default:
          scripts[key] = e.value;
      }
    });

    return scripts;
  }

  static get defaultOptions() {
    return {
      actions: [],
      keywords: []
    };
  }

  static matchesFileName(name) {
    return name.match(/^package\.json$/);
  }

  optionalDevModules(modules = new Set()) {
    return new Set(['cracks', 'dont-crack'].filter(m => modules.has(m)));
  }

  async usedDevModules(content) {
    content = await content;

    const pkg = content.length === 0 ? {} : JSON.parse(content);

    return moduleNames(pkg.release);
  }

  /**
   * Deliver some key properties
   * @param {Branch} branch
   * @return {Object}
   */
  async properties(branch) {
    try {
      const content = await branch.content(this.path);
      const pkg = JSON.parse(content.content);

      const properties = {
        npm: { name: pkg.name, fullName: pkg.name }
      };

      if (pkg.name !== undefined) {
        const m = pkg.name.match(/^(\@[^\/]+)\/(.*)/);
        if (m) {
          properties.npm.organization = m[1];
          properties.npm.name = m[2];
        }
      }

      if (pkg.template !== undefined && pkg.template.repository !== undefined) {
        properties.templateRepo = pkg.template.repository.url;
      }

      ['description', 'name', 'module', 'browser', 'version'].forEach(key => {
        if (pkg[key] !== undefined && pkg[key] !== `{{${key}}}`) {
          if (!(key === 'version' && pkg[key] === '0.0.0-semantic-release')) {
            properties[key] = pkg[key];
          }
        }
      });

      return properties;
    } catch (e) {}

    return {};
  }

  async mergeContent(context, original, templateContent) {
    const originalLastChar = original[original.length - 1];
    const originalTemplate = JSON.parse(templateContent);

    const targetRepository = context.targetBranch.repository;

    let target =
      original === undefined || original === '' ? {} : JSON.parse(original);

    const template = Object.assign({}, originalTemplate, {
      repository: {
        type: targetRepository.type,
        url: targetRepository.url
      },
      bugs: {
        url: context.targetBranch.issuesURL
      },
      homepage: context.targetBranch.homePageURL,
      template: {
        repository: {
          url: context.templateBranch.url
        }
      }
    });
    template.template = Object.assign({}, target.template, template.template);

    let messages = [];
    const properties = context.properties;

    if (target.name === undefined || target.name === '') {
      const m = targetRepository.name.match(/^([^\/]+)\/(.*)/);
      target.name = m ? m[2] : context.targetBranch.name;
    }

    if (target.module !== undefined && !target.module.match(/\{\{module\}\}/)) {
      properties.module = target.module;
    }

    properties.main =
      target.main && !target.main.match(/\{\{main\}\}/)
        ? target.main
        : 'dist/index.js';

    const slots = {
      repository: 'chore(package): correct repository url',
      bugs: 'chore(package): set bugs url from template',
      homepage: 'chore(package): homepage from template',
      template: 'chore(package): set template repo'
    };
    Object.keys(slots).forEach(key => {
      const templateValue = template[key];
      const d = diff(target[key], templateValue);

      if (
        templateValue !== undefined &&
        d.length > 0 &&
        !(
          d[0].type === 'add' &&
          d[0].oldValue === undefined &&
          d[0].newValue === undefined
        )
      ) {
        console.log(d);
        messages.push(slots[key]);
        target[key] = templateValue;
      }
    });

    const decodedScripts = Package.decodeScripts(target.scripts);

    const usedDevModules = await context.usedDevModules();

    //context.debug(`usedDevModules: ${Array.from(usedDevModules).join(',')}`);

    const optionalDevModules = context.optionalDevModules(usedDevModules);

    /*context.debug(
      `optionalDevModules: ${Array.from(optionalDevModules).join(',')}`
    );*/

    const deepProperties = {
      devDependencies: { type: 'chore', scope: 'package', merge: defaultMerge },
      dependencies: { type: 'fix', scope: 'package', merge: defaultMerge },
      peerDependencies: { type: 'fix', scope: 'package', merge: defaultMerge },
      optionalDependencies: {
        type: 'fix',
        scope: 'package',
        merge: defaultMerge
      },
      bundeledDependencies: {
        type: 'fix',
        scope: 'package',
        merge: defaultMerge
      },
      scripts: { type: 'chore', scope: 'scripts', merge: defaultMerge },
      engines: { type: 'chore', scope: 'engines', merge: defaultMerge }
    };

    Object.keys(deepProperties).forEach(
      name => (deepProperties[name].name = name)
    );

    Object.keys(deepProperties).forEach(category => {
      if (template[category] !== undefined) {
        Object.keys(template[category]).forEach(d => {
          if (target[category] === undefined) {
            target[category] = {};
          }

          const tp = context.expand(template[category][d]);
          if (
            category === 'devDependencies' &&
            target.dependencies !== undefined &&
            target.dependencies[d] === tp
          ) ; else {
            deepProperties[category].merge(
              target[category],
              target[category][d],
              tp,
              deepProperties[category],
              d,
              messages
            );
          }
        });
      }
    });

    Object.keys(template).forEach(p => {
      if (target[p] === undefined) {
        target[p] = template[p];
        messages.push(`chore(package): add ${p} from template`);
      }
    });

    target.scripts = Package.encodeScripts(
      Package.mergeScripts(
        Package.decodeScripts(template.scripts),
        decodedScripts
      )
    );

    if (target.module === '{{module}}') {
      delete target.module;
    }

    if (
      target.contributors !== undefined &&
      target.author !== undefined &&
      target.author.name !== undefined
    ) {
      const m = target.author.name.match(/(^[^<]+)<([^>]+)>/);
      if (m !== undefined) {
        const name = String(m[1]).replace(/^\s+|\s+$/g, '');
        const email = m[2];

        if (
          target.contributors.find(c => c.name === name && c.email === email)
        ) {
          delete target.author;
        }
      }
    }

    const toBeDeletedModules =
      target.devDependencies === undefined
        ? []
        : Array.from(
            context.optionalDevModules(
              new Set(Object.keys(target.devDependencies))
            )
          ).filter(m => !usedDevModules.has(m));

    toBeDeletedModules.forEach(d => {
      messages = messages.filter(
        m => !m.startsWith(`chore(package): add ${d}@`)
      );
      delete target.devDependencies[d];
    });

    target = deleter(target, template, messages, []);

    if (target.keywords !== undefined) {
      delete target.keywords['npm-package-template'];
    }

    Object.keys(this.options.keywords).forEach(r =>
      addKeyword(target, new RegExp(r), this.options.keywords[r], messages)
    );

    removeKeyword(target, ['null', null, undefined], messages);

    this.options.actions.forEach(action => {
      if (action.op === 'replace') {
        const v = jsonpath.value(template, action.path);
        if (v !== undefined) {
          const oldValue = jsonpath.value(target, action.path);
          if (oldValue !== v) {
            jsonpath.value(target, action.path, v);
            messages.push(
              `chore(package): set ${action.path}='${v}' as in template`
            );
          }
        }
      }
    });

    target = context.expand(target);
    const sortedTarget = normalizePackage(target);

    let newContent = JSON.stringify(sortedTarget, undefined, 2);
    const lastChar = newContent[newContent.length - 1];

    // keep trailing newline
    if (originalLastChar === '\n' && lastChar === '}') {
      newContent += '\n';
    }

    const changed = original !== newContent;

    if (changed && messages.length === 0) {
      messages.push('chore(package): update package.json from template');
    }

    return {
      content: newContent,
      messages,
      changed
    };
  }
}

function deleter(object, reference, messages, path) {
  if (
    typeof object === 'string' ||
    object instanceof String ||
    object === true ||
    object === false ||
    object === undefined ||
    object === null ||
    typeof object === 'number' ||
    object instanceof Number
  ) {
    return object;
  }

  if (Array.isArray(object)) {
    return object.map((e, i) => {
      path.push(i);
      const n = deleter(
        object[i],
        Array.isArray(reference) ? reference[i] : undefined,
        messages,
        path
      );
      path.pop();
      return n;
    });
  }

  if (reference) {
    Object.keys(reference).forEach(key => {
      path.push(key);

      if (reference[key] === '--delete--' && object[key] !== undefined) {
        if (object[key] !== '--delete--') {
          messages.push(`chore(npm): delete ${path.join('.')}`);
        }
        delete object[key];
      } else {
        object[key] = deleter(object[key], reference[key], messages, path);
      }
      path.pop();
    });
  }

  return object;
}

function removeKeyword(pkg, keywords, messages) {
  if (pkg.keywords !== undefined) {
    keywords.forEach(keyword => {
      if (pkg.keywords.find(k => k === keyword)) {
        messages.push(`docs(package): remove keyword ${keyword}`);
        pkg.keywords = pkg.keywords.filter(k => k !== keyword);
      }
    });

    if (
      (pkg.keywords.length === 1 && pkg.keywords[0] === null) ||
      pkg.keywords[0] === undefined
    ) {
      messages.push(`docs(package): remove keyword null`);
      pkg.keywords = [];
    }
  }
}

function addKeyword(pkg, regex, keyword, messages) {
  if (keyword === undefined || keyword === null || keyword === 'null') {
    return;
  }

  if (pkg.name.match(regex)) {
    if (pkg.keywords === undefined) {
      pkg.keywords = [];
    }
    if (!pkg.keywords.find(k => k === keyword)) {
      messages.push(`docs(package): add keyword ${keyword}`);
      pkg.keywords.push(keyword);
    }
  }
}

function getVersion(e) {
  const m = e.match(/([\d\.]+)/);
  return m ? Number(m[1]) : undefined;
}

function normalizeVersion(e) {
  return e.replace(/^[\^\$]/, '');
}

/**
 *
 */
function defaultMerge(destination, target, template, dp, name, messages) {
  if (template === '-') {
    if (target !== undefined) {
      messages.push(`${dp.type}(${dp.scope}): remove ${name}@${target}`);
      delete destination[name];
    }

    return;
  }

  if (target === undefined) {
    messages.push(`${dp.type}(${dp.scope}): add ${name}@${template}`);
    destination[name] = template;
  } else if (template !== target) {
    if (dp.name === 'engines') {
      if (getVersion(target) > getVersion(template)) {
        return;
      }
    }
    if (dp.name === 'devDependencies') {
      if (
        diffVersion(normalizeVersion(target), normalizeVersion(template)) >= 0
      ) {
        return;
      }
    }

    messages.push(`${dp.type}(${dp.scope}): ${name}@${template}`);

    destination[name] = template;
  }
}

/**
 * bring package into nomalized (sorted) form
 * @param {Object} source
 * @return {Object} normalized source
 */
function normalizePackage(source) {
  const normalized = {};

  sortedKeys.forEach(key => {
    if (source[key] !== undefined) {
      normalized[key] = source[key];
    }
  });

  Object.keys(source).forEach(key => {
    if (normalized[key] === undefined) {
      normalized[key] = source[key];
    }
  });

  return normalized;
}

const recast = require('recast');
const babylon = require('recast/parsers/babylon');

class Rollup extends File {
  static matchesFileName(name) {
    return name.match(/rollup\.config\.js/);
  }

  get needsTemplate() {
    return false;
  }

  optionalDevModules(modules) {
    return new Set(
      Array.from(modules).filter(
        m => m.match(/rollup-plugin/) || m.match(/babel-preset/)
      )
    );
  }

  usedDevModules(content) {
    const modules = new Set();

    const ast = recast.parse(content, {
      parser: babylon
    });

    for (const decl of ast.program.body) {
      if (decl.type === 'ImportDeclaration') {
        modules.add(decl.source.value);
      }
    }

    return modules;
  }

  async mergeContent(context, original, template) {
    if (template === '') {
      return {
        content: original,
        changed: false
      };
    }
    if (original === '') {
      return {
        content: template,
        messages: ['chore(rollup): copy from template'],
        changed: true
      };
    }

    const messages = [];

    try {
      const recastOptions = {
        parser: babylon
      };

      const templateAST = recast.parse(template, recastOptions);
      const ast = recast.parse(original, recastOptions);

      removeUseStrict(ast);

      const exp = exportDefaultDeclaration(ast);
      const templateExp = exportDefaultDeclaration(templateAST);

      if (exp !== undefined && exp.properties !== undefined) {
        let output, dest;

        const banner = removePropertiesKey(exp.properties, 'banner');

        for (const p of exp.properties) {
          switch (p.key.name) {
            case 'targets':
              dest = p.value.elements[0].properties[0]; //.find(x => x.name === 'dest');
              const op = findProperty(templateExp.properties, 'output');
              if (op !== undefined) {
                p.key.name = 'output';
                p.value = op.value;
                output = p;
              }
              break;
            case 'entry':
              const ip = findProperty(templateExp.properties, 'input');
              if (ip !== undefined) {
                p.key.name = 'input';
                p.value = ip.value;
              }
          }
        }

        if (findProperty(exp.properties, 'input') === undefined) {
          exp.properties.push(findProperty(templateExp.properties, 'input'));
        }

        const originalOutput = findProperty(exp.properties, 'output');
        const templateOutput = findProperty(templateExp.properties, 'output');

        if (originalOutput === undefined) {
          exp.properties.push(templateOutput);
        } else {
          mergeKeys(
            templateOutput,
            originalOutput,
            [
              'format',
              'file',
              'dir',
              'name',
              'globals',
              'paths',
              'banner',
              'footer',
              'intro',
              'outro',
              'sourcemap',
              'sourcemapFile',
              'interop',
              'extend',
              'exports',
              'amd',
              'indent',
              'strict',
              'freeze',
              'legacy',
              'namespaceToStringTag'
            ],
            messages
          );

          if (output !== undefined) {
            if (banner !== undefined) {
              output.value.properties.push(banner);
            }

            if (dest !== undefined) {
              const file = findProperty(output.value.properties, 'file');
              if (file !== undefined) {
                file.value = dest.value;
              }
            }
          }

          removePropertiesKey(exp.properties, 'format');
          removePropertiesKey(exp.properties, 'sourceMap');
          removePropertiesKey(exp.properties, 'dest');
        }
      }
      const originalImports = importDeclarationsByLocalName(ast);
      const templateImports = importDeclarationsByLocalName(templateAST);

      const addedImports = [];

      templateImports.forEach((value, key) => {
        if (originalImports.get(key) === undefined) {
          ast.program.body = [value, ...ast.program.body];
          addedImports.push(key);
        }
      });

      if (addedImports.length > 0) {
        messages.push(`chore(rollup): import ${addedImports.join(',')}`);
      }

      const addedPlugins = [];
      const originalPlugins = pluginsFromExpression(exp);
      const templatePlugins = pluginsFromExpression(templateExp);

      templatePlugins.forEach(templatePlugin => {
        if (
          originalPlugins.find(
            op => op.callee.name === templatePlugin.callee.name
          ) === undefined
        ) {
          originalPlugins.push(templatePlugin);
          addedPlugins.push(templatePlugin.callee.name);
        }
      });

      if (addedPlugins.length > 0) {
        messages.push(`chore(rollup): add ${addedPlugins.join(',')}`);
      }

      const content = recast.print(ast).code;
      const changed = content !== original;

      if (changed && messages.length === 0) {
        messages.push('chore(rollup): update from template');
      }

      return {
        content,
        changed,
        messages
      };
    } catch (e) {
      console.log(e);
      context.warn(`unable to parse ${this.path}`);
      context.error(e);
    }

    return {
      content: original,
      changed: false
    };
  }
}

function removeUseStrict(ast) {
  for (const i in ast.program.body) {
    const decl = ast.program.body[i];
    if (
      decl.type === 'ExpressionStatement' &&
      decl.expression.type === 'Literal' &&
      decl.expression.value === 'use strict'
    ) {
      ast.program.body.splice(i, 1);
      return;
    }
  }
}

function importDeclarationsByLocalName(ast) {
  const declarations = new Map();

  for (const decl of ast.program.body) {
    if (decl.type === 'ImportDeclaration') {
      declarations.set(decl.specifiers[0].local.name, decl);
    }
  }

  return declarations;
}

function exportDefaultDeclaration(ast) {
  for (const decl of ast.program.body) {
    if (decl.type === 'ExportDefaultDeclaration') {
      return decl.declaration;
    }
  }

  return undefined;
}

function pluginsFromExpression(exp) {
  if (exp.properties !== undefined) {
    const plugins = exp.properties.find(
      p => p !== undefined && p.key.name === 'plugins'
    );

    if (plugins !== undefined) {
      return plugins.value.elements;
    }
  }

  return [];
}

function findProperty(properties, name) {
  if (properties !== undefined) {
    const prop = properties.find(
      prop => prop !== undefined && prop.key.name === name
    );

    return prop;
  }
  return undefined;
}

function removePropertiesKey(properties, name) {
  if (properties === undefined) {
    return undefined;
  }

  const toBeRemoved = properties.findIndex(x => x && x.key.name === name);
  if (toBeRemoved >= 0) {
    const slot = properties[toBeRemoved];
    properties.splice(toBeRemoved, 1);
    return slot;
  }

  return undefined;
}

function mergeKeys(source, dest, knownKeys, messages) {
  const mergedKeys = [];

  if (source !== undefined) {
    knownKeys.forEach(key => {
      const destProp = findProperty(dest.value.properties, key);
      const sourceProp = findProperty(source.value.properties, key);
      if (sourceProp !== undefined && destProp === undefined) {
        mergedKeys.push(key);
        dest.value.properties.push(sourceProp);
      }
    });
  }

  if (mergedKeys.length > 0) {
    messages.push(`chore(rollup): add to output ${mergedKeys.join(',')}`);
  }

  return mergedKeys;
}

class License extends File {
  static matchesFileName(name) {
    return name.match(/^LICENSE/);
  }

  async mergeContent(context, original, template) {
    const properties = context.properties;
    const messages = [];

    let year = properties['date.year'];

    const m = original.match(
      /opyright\s*\(c\)\s*(\d+)([,\-\d]+)*(\s*(,|by)\s*(.*))?/
    );

    if (m) {
      const years = new Set();
      years.add(parseInt(m[1], 10));

      if (m[2] !== undefined) {
        m[2].split(/\s*[,\-]\s*/).forEach(y => {
          if (y.length > 0) {
            years.add(parseInt(y, 10));
          }
        });
      }

      if (m[4] !== undefined) {
        properties['license.owner'] = m[5];
      }

      if (!years.has(year)) {
        years.add(year);
        messages.push(`chore(license): add current year ${year}`);
      }

      year = Array.from(years)
        .sort((a, b) => (a < b ? -1 : a > b ? 1 : 0))
        .join(',');
    }

    if (messages.length === 0) {
      messages.push('chore(license): update');
    }

    if (original !== '') {
      const content = original.replace(
        /opyright\s*\(c\)\s*(\d+)([,\-\d])*/,
        `opyright (c) ${year}`
      );

      return {
        changed: content !== original,
        messages,
        content
      };
    }

    return {
      content: context.expand(template),
      changed: true,
      messages: [`chore(license): add LICENSE`]
    };
  }
}

/**
 * File where every line is a key
 */
class MergeLineSet extends File {
  /**
   * entries to be skipped from result
   * @return {Set<string>}
   */
  get defaultIgnoreSet() {
    return new Set();
  }

  async mergeContent(context, original, template) {
    const result = new Set(template.split(/\n/));
    original.split(/\r?\n/).forEach(line => result.add(line));

    this.defaultIgnoreSet.forEach(entry => result.remove(entry));

    const content = Array.from(result.values()).join('\n');

    return {
      content,
      changed: content !== original,
      messages: [this.options.message || 'fix: updated from template']
    };
  }
}

/**
 *
 */
class MergeAndRemoveLineSet extends MergeLineSet {
  async mergeContent(context, original, template) {
    const toBeRemoved = new Set([
      ...template
        .split(/\r?\n/)
        .filter(l => l.startsWith('- '))
        .map(l => l.replace(/^-\s+/, '')),
      ...Array.from(this.defaultIgnoreSet)
    ]);

    const result = new Set(
      template.split(/\r?\n/).filter(l => !l.match(/^-\s+/))
    );
    original
      .split(/\r?\n/)
      .filter(l => !toBeRemoved.has(l))
      .forEach(line => result.add(line));

    const content = Array.from(result.values()).join('\n');

    return {
      content,
      changed: content !== original,
      messages: [
        this.options.message || `chore: ${this.path} updated from template`
      ]
    };
  }
}

/**
 *
 */
class NpmIgnore extends MergeAndRemoveLineSet {
  static matchesFileName(name) {
    return name === '.npmignore';
  }

  /**
   * entries to be skipped from result
   * @see https://docs.npmjs.com/misc/developers
   * @return {Set<string>}
   */
  get defaultIgnoreSet() {
    return new Set([
      '.*.swp',
      '._*',
      '.DS_Store',
      '.git',
      '.github',
      '.hg',
      '.npmrc',
      '.lock-wscript',
      '.svn',
      '.wafpickle-*',
      'config.gypi',
      'CVS',
      'npm-debug.log'
    ]);
  }
}

/**
 * Overwrites none existing file from template
 */
class ReplaceIfEmpty extends File {
  get needsTemplate() {
    return false;
  }

  async mergeContent(context, original, template) {
    return original === ''
      ? {
          content: context.expand(template),
          changed: template !== '',
          messages: [
            this.options.message ||
              `chore: add missing ${this.path} from template`
          ]
        }
      : {
          path: this.path,
          content: original,
          changed: false
        };
  }
}

/**
 * Replace file from template (always)
 */
class Replace extends File {
  async mergeContent(context, original, template) {
    const content = context.expand(template);

    return {
      content,
      changed: content !== original,
      messages: [`chore: ${this.path} overwritten from template`]
    };
  }
}

class JSONFile extends File {
  static matchesFileName(name) {
    return name.match(/\.json$/);
  }

  get needsTemplate() {
    return false;
  }

  async mergeContent(context, original, templateRaw) {
    if (templateRaw === '' || templateRaw === undefined) {
      return undefined;
    }

    const target =
      original === '' || original === undefined ? {} : JSON.parse(original);
    const template = JSON.parse(templateRaw);

    Object.assign(target, template);

    const content = JSON.stringify(context.expand(target), undefined, 2);

    return {
      content,
      changed: content !== original,
      messages: ['chore: update from template']
    };
  }
}

class JSDoc extends JSONFile {
  optionalDevModules(modules) {
    return new Set(Array.from(modules).filter(m => m.match(/babel-preset/)));
  }

  usedDevModules(content) {
    const modules = new Set();

    const json = JSON.parse(content);

    if (json.babel !== undefined && json.babel.presets !== undefined) {
      json.babel.presets.forEach(m => modules.add(`babel-preset-${m}`));
    }

    return modules;
  }
}

/**
 * context prepared to execute one package
 * @param {Context} context
 * @param {string} targetBranchName
 *
 * @property {Object} ctx
 * @property {Map<string,Object>} files
 */
class PreparedContext$$1 {
  static get mergers() {
    return [
      Rollup,
      Travis,
      Readme,
      Package,
      JSONFile,
      JSDoc,
      Travis,
      MergeAndRemoveLineSet,
      MergeLineSet,
      NpmIgnore,
      License,
      ReplaceIfEmpty,
      Replace
    ];
  }

  static async from(context, targetBranchName) {
    const pc = new PreparedContext$$1(context, targetBranchName);
    await pc.initialize();
    return pc;
  }

  static async execute(context, targetBranchName) {
    const pc = new PreparedContext$$1(context, targetBranchName);
    await pc.initialize();
    return pc.execute();
  }

  constructor(context, targetBranchName) {
    Object.defineProperties(this, {
      ctx: {
        value: expressionExpander.createContext({
          properties: Object.assign({}, context.properties),
          keepUndefinedValues: true,
          leftMarker: '{{',
          rightMarker: '}}',
          markerRegexp: '{{([^}]+)}}',
          evaluate: (expression, context, path) =>
            jsonpath.value(this.properties, expression)
        })
      },
      files: {
        value: new Map()
      },
      context: { value: context },
      targetBranchName: { value: targetBranchName }
    });
  }

  get mergers() {
    return this.constructor.mergers;
  }

  get provider() {
    return this.context.provider;
  }

  get templateBranchName() {
    return this.context.templateBranchName;
  }

  get properties() {
    return this.ctx.properties;
  }

  expand(...args) {
    return this.ctx.expand(...args);
  }

  error(...args) {
    return this.context.error(...args);
  }

  warn(...args) {
    return this.context.warn(...args);
  }

  debug(...args) {
    return this.context.debug(...args);
  }

  info(...args) {
    return this.context.info(...args);
  }

  async initialize() {
    const context = this.context;
    const targetBranch = await context.provider.branch(this.targetBranchName);

    if (targetBranch.provider.name === 'GithubProvider') {
      this.properties.github = {
        user: targetBranch.repository.owner,
        repo: targetBranch.repository.condensedName
      };
    }

    const pkg = new Package('package.json');

    Object.assign(this.properties, await pkg.properties(targetBranch));

    let templateBranch;

    if (context.templateBranchName === undefined) {
      try {
        templateBranch = await context.provider.branch(
          this.properties.templateRepo
        );
      } catch (e) {}

      if (templateBranch === undefined) {
        throw new Error(
          `Unable to extract template repo url from ${targetBranch.name} ${
            pkg.path
          }`
        );
      }
    } else {
      templateBranch = await context.provider.branch(this.templateBranchName);
    }

    Object.defineProperties(this, {
      templateBranch: { value: templateBranch },
      targetBranch: { value: targetBranch }
    });
  }

  static async createFiles(branch, mapping = Context.defaultMapping) {
    const files = await branch.list();

    let alreadyPresent = new Set();

    return mapping
      .map(m => {
        const found = micromatch(
          files.filter(f => f.type === 'blob').map(f => f.path),
          m.pattern
        );

        const notAlreadyProcessed = found.filter(f => !alreadyPresent.has(f));

        alreadyPresent = new Set([...Array.from(alreadyPresent), ...found]);

        return notAlreadyProcessed.map(f => {
          const merger =
            this.mergers.find(merger => merger.name === m.merger) ||
            ReplaceIfEmpty;
          return new merger(f, m.options);
        });
      })
      .reduce((last, current) => Array.from([...last, ...current]), []);
  }

  addFile(file) {
    this.files.set(file.path, file);
  }

  /**
   * all used dev modules
   * @return {Set<string>}
   */
  async usedDevModules() {
    const usedModuleSets = await Promise.all(
      Array.from(this.files.values()).map(async file => {
        if (file.path === 'package.json') {
          return file.usedDevModules(
            file.targetContent(this, { ignoreMissing: true })
          );
        } else {
          const m = await file.merge(this);
          return file.usedDevModules(m.content);
        }
      })
    );

    return usedModuleSets.reduce(
      (sum, current) => new Set([...sum, ...current]),
      new Set()
    );
  }

  optionalDevModules(modules) {
    return Array.from(this.files.values())
      .map(file => file.optionalDevModules(modules))
      .reduce((sum, current) => new Set([...sum, ...current]), new Set());
  }

  async trackUsedModule(targetBranch) {
    const templateBranch = this.templateBranch;

    let templatePullRequest;
    let newTemplatePullRequest = false;
    let templatePRBranch = await templateBranch.repository.branch(
      'template-add-used-1'
    );

    const pkg = new Package('package.json');

    const templatePackage = await (templatePRBranch
      ? templatePRBranch
      : templateBranch
    ).content(pkg.path, { ignoreMissing: true });

    const templatePackageContent = templatePackage.content;

    const templatePackageJson =
      templatePackageContent === undefined || templatePackageContent === ''
        ? {}
        : JSON.parse(templatePackageContent);

    if (this.context.trackUsedByModule) {
      const name = targetBranch.fullCondensedName;

      if (templatePackageJson.template === undefined) {
        templatePackageJson.template = {};
      }
      if (!Array.isArray(templatePackageJson.template.usedBy)) {
        templatePackageJson.template.usedBy = [];
      }

      if (!templatePackageJson.template.usedBy.find(n => n === name)) {
        templatePackageJson.template.usedBy.push(name);
        templatePackageJson.template.usedBy = templatePackageJson.template.usedBy.sort();

        if (templatePRBranch === undefined) {
          templatePRBranch = await templateBranch.repository.createBranch(
            'template-add-used-1',
            templateBranch
          );
          newTemplatePullRequest = true;
        }

        await templatePRBranch.commit(`fix: add ${name}`, [
          {
            path: 'package.json',
            content: JSON.stringify(templatePackageJson, undefined, 2)
          }
        ]);

        if (newTemplatePullRequest) {
          templatePullRequest = await templateBranch.createPullRequest(
            templatePRBranch,
            {
              title: `add ${name}`,
              body: `add tracking info for ${name}`
            }
          );
        }
      }
    }

    return { templatePackageJson, templatePRBranch, templatePullRequest };
  }

  /**
   * @return {Promise<PullRequest>}
   */
  async execute() {
    const templateBranch = this.templateBranch;
    const targetBranch = this.targetBranch;

    const {
      templatePackageJson,
      templatePRBranch,
      templatePullRequest
    } = await this.trackUsedModule(targetBranch);

    const files = await PreparedContext$$1.createFiles(
      templateBranch,
      templatePackageJson.template && templatePackageJson.template.files
    );

    files.forEach(f => this.addFile(f));

    const merges = (await Promise.all(
      files.map(async f => f.merge(this))
    )).filter(m => m !== undefined && m.changed);

    if (merges.length === 0) {
      this.info(`${targetBranch.fullCondensedName}: -`);
      return;
    }

    this.info(
      merges
        .map(m => `${targetBranch.fullCondensedName}: ${m.messages[0]}`)
        .join(',')
    );

    if (this.dry) {
      this.info(`${targetBranch.fullCondensedName}: dry run`);
      return;
    }

    let newPullRequestRequired = false;
    const prBranchName = 'template-sync-1';
    let prBranch = (await this.targetBranch.repository.branches()).get(
      prBranchName
    );

    if (prBranch === undefined) {
      newPullRequestRequired = true;
      prBranch = await this.targetBranch.repository.createBranch(
        prBranchName,
        targetBranch
      );
    }

    const messages = merges.reduce((result, merge) => {
      merge.messages.forEach(m => result.push(m));
      return result;
    }, []);

    await prBranch.commit(messages.join('\n'), merges);

    if (newPullRequestRequired) {
      try {
        const pullRequest = await targetBranch.createPullRequest(prBranch, {
          title: `merge package from ${templateBranch.fullCondensedName}`,
          body: merges
            .map(
              m =>
                `${m.path}
---
- ${m.messages.join('\n- ')}
`
            )
            .join('\n')
        });
        this.info(`${targetBranch.fullCondensedName}: ${pullRequest.name}`);

        return pullRequest;
      } catch (err) {
        this.error(err);
      }
    } else {
      const pullRequest = new targetBranch.provider.pullRequestClass(
        targetBranch.repository,
        'old'
      );

      this.info(
        `${targetBranch.fullCondensedName}: update PR ${pullRequest.name}`
      );
      return pullRequest;
    }
  }
}

/**
 * @param {RepositoryProvider} provider
 * @param {Object} options
 *
 * @property {RepositoryProvider} provider
 * @property {Object} options
 * @property {string} options.templateBranchName
 */
class Context {
  static get defaultMapping() {
    return [
      { merger: 'Package', pattern: '**/package.json' },
      { merger: 'Travis', pattern: '.travis.yml' },
      { merger: 'Readme', pattern: '**/README.*' },
      { merger: 'JSDoc', pattern: '**/jsdoc.json' },
      { merger: 'Rollup', pattern: '**/rollup.config.js' },
      { merger: 'License', pattern: 'LICENSE' },
      {
        merger: 'MergeAndRemoveLineSet',
        pattern: '.gitignore',
        options: { message: 'chore(git): update {{path}} from template' }
      },
      {
        merger: 'NpmIgnore',
        pattern: '.npmignore',
        options: { message: 'chore(npm): update {{path}} from template' }
      },
      { merger: 'ReplaceIfEmpty', pattern: '**/*' }
    ];
  }

  constructor(provider, options) {
    options = Object.assign(
      {},
      {
        logger: console,
        dry: false,
        trackUsedByModule: false
      },
      options
    );

    options.properties = Object.assign(
      {
        'date.year': new Date().getFullYear()
      },
      options.properties
    );

    Object.defineProperties(this, {
      trackUsedByModule: {
        value: options.trackUsedByModule
      },
      dry: {
        value: options.dry
      },
      logger: {
        value: options.logger
      },
      provider: {
        value: provider
      },
      properties: {
        value: options.properties
      },
      templateBranchName: {
        value: options.templateBranchName
      }
    });
  }

  get defaultMapping() {
    return this.constructor.defaultMapping;
  }

  debug(...args) {
    this.logger.debug(...args);
  }

  warn(...args) {
    this.logger.warn(...args);
  }

  info(...args) {
    this.logger.info(...args);
  }

  error(...args) {
    this.logger.error(...args);
  }
}

exports.PreparedContext = PreparedContext$$1;
exports.Context = Context;
